%
% The MaximumTransferRate.m test the maximum transfer rate you can achieve
%

clear all;

% create a timetagger instance
tagger = TimeTagger;
tagger.reset();

disp('*******************************');
disp('*** Turn on the test signal ***');
disp('*******************************');

disp('The data transfer rate will be measured by turning on the internal test signal')
disp('on several channels, with which in total the USB transfer rate is exceeded.')
disp('By measuring the rate of incoming time tags, the maximum data rate is measured.')

disp('')
disp('Please close all other programs to get the maximum data rate possible.')
disp('')

if strcmp(tagger.getModel(), 'Time Tagger Ultra')
    % Time Tagger Ultra
    % set test signal rate: 25 MHz
    tagger.setTestSignalDivider(2);
    allRisingChannels = tagger.getChannelList(1); %get all channel numbers of the rising edges
    testChannels = allRisingChannels(1:8); % select the first 8 channels
    tagger.setTestSignal(testChannels, true)
    disp('Total rate: 8x 25 MHz');
else
    % Time Tagger 20
    % test signal rate ~0.8 MHz
    % apply test signal to all channels
    testChannels = tagger.getChannelList(0);
    disp('Total rate: 16x 0.8 MHz');
end
tagger.setTestSignal(testChannels, true)

disp(' ');
disp('******************************************************');
disp('*** Capture the stream via the TimeTagStream class ***');
disp('******************************************************');
disp(' ');
disp('The TimeTagStream class stores the incoming tags in memory');
disp('and is therefore the fastest way to access the time tag stream.');
disp(' ');
bufferSize = 10e6; % tags
capture_time = 0.1; %s
averages = 100;
disp(['Capture buffer size: ' num2str(bufferSize/1e6)   ' MTags']);
disp(['Capture duration:    ' num2str(capture_time) ' s']);
disp(['Test runs:           ' num2str(averages)]);
disp(' ');

totalCounts = 0;
totalTime = 0;

stream = TTTimeTagStream(tagger, bufferSize, testChannels);

transfer_rate = zeros(1, averages);

tic
startTime = toc;
endTime = startTime;
for i = 1:averages
    pause(capture_time - (toc - endTime))
    endTime = toc;
    data = stream.getData();
    dt = endTime - startTime;
    startTime = endTime;
    totalCounts = totalCounts + data.size;
    totalTime = totalTime + dt;
    transfer_rate(i) = double(data.size) / 1e6 / dt;
    fprintf('#%3d: %5.1f MTags/s | dt: %7.3f ms | buffer fill: %3.0f %%\n', i, transfer_rate(i), dt * 1e3, data.size * 100 / bufferSize);
end

transfer_rate = sort(transfer_rate);

fprintf('\n>>>  Average transfer rate: %4.1f MTags/s  <<<\n', totalCounts / 1e6 / totalTime);

if strcmp(tagger.getModel(), 'Time Tagger Ultra')
    disp('The maximum transfer rate is not limited by USB 3.0 but by the single thread CPU performance.');
    disp('A very good processor (i9-9900K, max. 5 GHz) can process in total 65 MTags/s if 8 channels are used.');
    disp('If case the transfer rate dropping during the measurement compared the the very first measurement points,');
    disp('your CPU is probably throttling. This happens especially with mobile processors.');
    disp('In case your data rate is only between 10-12 MTags/s, it is likely that you have connected the Time Tagger')
    disp('to a USB 2.0 port, instead of a USB 3.0 port.');
else
    disp('The transfer rate should exceed 8 MTags/s and is limited by the USB 2.0 bandwith.');
    disp('Usually, the CPU perfomance is not playing a role at these data rates.');
end

stream.delete();
tagger.freeTimeTagger();
clear tagger