/*
This file is part of Time Tagger software defined digital data acquisition.

Copyright (C) 2011-2016 Swabian Instruments
All Rights Reserved

Unauthorized copying of this file is strictly prohibited.
*/

#include "TimeTagger.h"
#include "CustomMeasurement.h"

#include <cstddef>

CustomMeasurement::CustomMeasurement(TimeTaggerBase* tagger, std::vector<channel_t> channels)
  : IteratorBase(tagger), channels(channels),  customData(channels.size()) {
  clear();

  // we have to tell the FPGA which channels are currently in use - so register them (tags of channels which are not used are not sent to the computer)
  for (auto & elem : channels)
    registerChannel(elem);

  finishInitialization();
}

CustomMeasurement::~CustomMeasurement() {
  stop();
}

// get data copies the local data to a newly allocated memory and the pointer is returned
void CustomMeasurement::getData(std::function<long long*(size_t)> array_out) {
  long long *arr = array_out(channels.size());

  auto lk = getLock();

  for (size_t i = 0; i < channels.size(); i++) {
    arr[i] = customData[i];
  }
}

void CustomMeasurement::clear_impl() {
  for (size_t c = 0; c < channels.size(); c++)
    customData[c] = 0;
}

void CustomMeasurement::on_start() {
  // optional callback
}

void CustomMeasurement::on_stop() {
  // optional callback
}

// here we handle the incoming time tags
bool CustomMeasurement::next_impl(std::vector<Tag>& incoming_tags, timestamp_t begin_time, timestamp_t end_time) {
  // iterate over all the tags recievied
  for (const Tag& tag : incoming_tags) {
    if (tag.type != Tag::Type::TimeTag) {
      // you always have to handle the overflows. 
      // here you must implement what should happen when FPGA data buffer is overflowing because of the limited transfer bandwidth
      // sometimes you have to restart the measurement - sometimes you're fine with the amount of missed events - sometimes you don't care about overflows. 
      // for this example we simply ignore the overflows
    } else {
      for (size_t c = 0; c < channels.size(); c++) {
        // the custom measurement here simply counts the number of incoming tags till the end
        if (tag.channel == channels[c]) {
          customData[c]++;
        }
      }
    }
  }

  // return if incoming_tags was modified. If so, please keep care about the requirements:
  // -- all tags must be sorted
  // -- begin_time <= tags < end_time
  return false;
}
