/*
This file is part of Time Tagger software defined digital data acquisition.

Copyright (C) 2011-2016 Swabian Instruments
All Rights Reserved

Unauthorized copying of this file is strictly prohibited.
*/

#include "TimeTagger.h"
#include "Iterators.h"

#include "CustomMeasurement.h"

#include <iostream>
#include <stdio.h>
#include <chrono>
#include <thread>

int main()
{
  std::cout << "Welcome to the Time Tagger API Demo\n";

  if (sizeof(void *) == 4)
  {
    std::cout << "You are using the 32bit version of the Time Tagger. Please switch to the 64bit version for superior performance.\n";
  }

  std::vector<std::string> taggers = scanTimeTagger();
  if (taggers.empty())
  {
    std::cout << "\nNo time tagger found.\nPlease attach a Time Tagger.";
    getchar();
    return 1;
  }

  // connect to a time tagger
  TimeTagger *t = createTimeTagger();

  // turn in the test signal on the first two channels for all measurements
  for (int i = 1; i < 3; i++)
  {
    t->setTestSignal(i, true);
  }

  // *************************** Counterrate class example
  {
    std::cout << "\nExample for the Counterrate class\n";
    std::vector<channel_t> v = {1, 2, 3, 4};
    Countrate c(t, v);

    std::this_thread::sleep_for(std::chrono::milliseconds(1000));

    // getData first allocates the memory and then copies the internal data to the new allocated one
    std::vector<double> data;
    c.getData([&data](size_t size) {data.resize(size); return data.data(); });
    std::cout << "Countrate of channel 1: " << data[0] << " (should be > 0 because the test signal is turned on)\n";
    std::cout << "Countrate of channel 4: " << data[3] << " (should be 0 because the test signal is off)\n";
  }

  {
    // *************************** Counter class example
    std::cout << "\nExample for the Counter class\n";
    std::vector<channel_t> v = {1, 2, 3, 4};
    const int bins = 10;
    Counter c(t, v, 1e8, 10); // sampling time per bin: 1e8 = 100ms, 10 bins

    std::this_thread::sleep_for(std::chrono::milliseconds(1000));

    // getData will always return a 1D object
    // this is due to the interoperability for the different programming languages
    std::vector<int> data;
    c.getData([&data](size_t size1, size_t size2) {data.resize(size1 * size2); return data.data(); });
    // reshape to 2d
    // data2d[channel][bin]
    int N = bins;
    int M = v.size();
    std::vector<int *> pointers;
    for (int i = 0; i < M; ++i)
    {
      pointers.push_back(&data[i * N]);
    }
    int **data2d = pointers.data();

    std::cout << "1st bin of channel 1: " << data2d[0][0] << " (should be > 0 because the test signal is turned on)\n";
    std::cout << "1st bin of channel 4: " << data2d[3][0] << " (should be 0 because the test signal is off)\n";
  }

  {
    // *************************** Custom implemented measurement
    std::cout << "\nExample for the CustomMeasurement class\n";
    std::vector<channel_t> v = {1, 2, 3, 4};
    CustomMeasurement c(t, v);

    std::this_thread::sleep_for(std::chrono::milliseconds(1000));

    std::vector<long long> data;
    c.getData([&data](size_t size) {data.resize(size); return data.data(); });

    std::cout << "channel 1: " << data[0] << " (should be > 0 because the test signal is turned on)\n";
    std::cout << "channel 4: " << data[3] << " (should be 0 because the test signal is off)\n";
  }

  freeTimeTagger(t);
  std::cout << "\n\nPress any key to return to Visual Studio.\n";
  getchar();

  return 0;
}
