﻿using System;
using SwabianInstruments.TimeTagger;
using System.Threading;
using System.Linq;

namespace TTCSharp
{
  class Program
  {
    static void Main(string[] args)
    {
      if (!Environment.Is64BitProcess)
      {
          Console.WriteLine("The C# code is is running in the 32-bit mode.");
          Console.WriteLine("For improved performance, please change the project settings such that it is running in the 64 bit mode.");
      }
      // This is a very basic example to get started with the Time Tagger in C#.

      // Implement custom logging
      TT.setLogger(delegate (int level, string message)
      {
        if (level >= (int) LogLevel.LOGGER_WARNING) {
          System.Console.WriteLine(message);
        }
      });

      TimeTagger t = TT.createTimeTagger();
      System.Console.WriteLine("Hello world! My serial is: " + t.getSerial() + "\n");

      // PART 1 - using the Correlation class
      t.reset();
      Console.WriteLine("PART 1 - Correlation class");
      Console.WriteLine("The internal test signal will be used to determine the jitter of the Time Tagger.");
      Console.WriteLine("The distribution returned is not around x=0 because the internal test signal\nhas a different delay than the external signal inputs.");

      // After the very basic hello world we use the scope class to get some data:
      using (Correlation s = new Correlation(t, channel_1: 1, channel_2: 2, binwidth: 50, n_bins: 1000))
      {
        t.setTestSignal(1, true);
        t.setTestSignal(2, true);
        t.sync();
        Thread.Sleep(1000);
        var data = s.getData();
        var x = s.getIndex();

        int maxValue = data.Max();
        int maxIndex = data.ToList().IndexOf(maxValue);

        for (int i = Math.Max(0,-10 + maxIndex); i < Math.Min(data.Length, 10 + maxIndex); i++)
        {
          System.Console.WriteLine(x[i] + " ps:  " + data[i] + " counts");
        }
      }

      // PART 2 - using the Scope class
      Console.WriteLine("\nPART 2 - Scope class");

      // After the very basic hello world we use the scope class to get some data:
      using (Scope s = new Scope(t, event_channels: new int[] { 1, 2 }, trigger_channel: 1, window_size: 100000000))
      {
        t.setTestSignal(1, true);
        t.setTestSignal(2, true);
        t.sync();
        Thread.Sleep(1000);
        var data = s.getData();
        System.Console.WriteLine("Number of events received: " + data[0].Length);
        for (int i = 0; i < data[0].Length; i++)
        {
          Event e = data[0][i];
          System.Console.WriteLine(e.time + " ps: " + e.state.ToString());
          if (i > 8)
          {
            System.Console.WriteLine("...\n");
            break;
          }
        }
      }

      // PART 2 - raw TimeTagStream access
      Console.WriteLine("PART 3 - raw TimeTagStream access");

      // An example for accessing the time tag stream
      var channels = new int[] { 1, 2, 3, 4 };
      t.setTestSignal(channels, true);
      uint bufferSize = 10000000;
      using (TimeTagStream tts = new TimeTagStream(t, bufferSize, channels))
      {

        // now we access the data stream 10 times for 0.1 second and do a simple (useless) evalutaion of the tags
        for (int i = 0; i < 10; i++)
        {
          Thread.Sleep(100);
          using (var mydata = tts.getData())
          {
            int evalChannel = 0;
            long evalTimestamp = 0;
            foreach (var number in mydata.getChannels())
            {
              evalChannel += number % 2;
            }
            foreach (var number in mydata.getTimestamps())
            {
              evalTimestamp += number % 2;
            }
            Console.WriteLine("Tags received " + mydata.size + "  hasOverflows " + mydata.hasOverflows + "  evalChannel " + evalChannel + "  evalTimeStamp " + evalTimestamp);
      	  }
        }

        System.Console.WriteLine("\nPress any key to return to visual studio.");
        System.Console.ReadKey();
      }
    }
  }
}
