"""
This example shows how to use FileWriter and FileReader for writing and reading
the time tag stream onto a disk.
"""

import tempfile
import os

from TimeTagger import createTimeTagger, FileWriter, FileReader
import numpy as np

from time import sleep

# create a timetagger instance
tagger = createTimeTagger()
tagger.reset()

print("*******************************************************")
print("*** Demonstrate how to dump the raw time tag stream ***")
print("*******************************************************")
print("")
print("Enabling test signal on channel 1 and 2.")
print("Frequency channel 1: ~800 kHz")
tagger.setTestSignal(1, True)
print("Frequency channel 2: ~800/2 kHz")
tagger.setTestSignal(2, True)
tagger.setEventDivider(2, 2)

print("")
tempdir = tempfile.TemporaryDirectory()
filepath = tempdir.name + os.sep + 'dump.ttbin'
print("1: Dump the data to " + filepath)
channels = [1, 2]

print("""
********************************************************
*  STEP 1: Write events from two channels into a file.
********************************************************
""")

print("""
Create a FileWriter object which saves the raw time tags of the given channels.
Parameters:
 (1) Time Tagger object the dump belongs to
 (2) file name for the raw time tag stream.
     FileWriter supports file splitting and more than one file will be created.
     Every created file will have the same filename plus a counter as a suffix.
 (3) channels which are dumped
""")

# The number of maximum tags which can be dumped is limited only by the size of your storage device.
file_writer = FileWriter(tagger, filepath, channels)
sleep(1)  # write for some time
file_writer.stop()

# Print statistics on file size and data-rate
print('Data recording complete.\n')
print('*************************** STATISTICS *****************************')
print('Recording duration:            {:8.0f} seconds'.format(
    file_writer.getCaptureDuration()*1e-12))
print('Events written to the file(s): {:8.0f} events'.format(
    file_writer.getTotalEvents()))
print('Bytes written to the file(s):  {:8.0f} bytes'.format(
    file_writer.getTotalSize()))
print('File writing data rate:        {:8.0f} bytes/s'.format(
    file_writer.getTotalSize()/(file_writer.getCaptureDuration()*1e-12)))
print('Compression efficiency:        {:8.2f} bytes/event'.format(
    file_writer.getTotalSize()/file_writer.getTotalEvents()))
print('NOTE: Compression efficiency depends on signal randomness.')
print('********************************************************************\n')

del file_writer

print("""
********************************************************
*  STEP 2: Read all tags data from the file(s).
********************************************************
""")

print('Read back the dumped stream from the filesystem using FileReader.')
file_reader = FileReader(filepath)

# now the data can be accessed via methods of the file_reader object

while file_reader.hasData():
    n_events = 100000  # Number of events to read at once
    # Read at most n_events.
    # data is an instance of TimeTagStreamBuffer
    data = file_reader.getData(n_events)
    print('Size of the returned data chunk: {:d} events\n'.format(data.size))

    print('Showing a few selected timetags')
    channel = data.getChannels()
    timestamps = data.getTimestamps()
    # TimeTag = 0, Error = 1, OverflowBegin = 2, OverflowEnd = 3, MissedEvents = 4
    overflow_types = data.getEventTypes()
    missed_events = data.getMissedEvents()

    eventName = ['0 (TimeTag)', '1 (Error)', '2 (OverflowBegin)',
                 '3 (OverflowEnd)', '4 (MissedEvents)']

    for i in range(3):
        print('\t TAG# {:6d} \t t = {:d} ps \t Channel: {:d} \t MissedEvents: {:d} \t EventType: {:s}'.format(
            i, timestamps[i], channel[i], missed_events[i], eventName[overflow_types[i]]))

    print('\t         ...')
    print('\t TAG# {:6d} \t t = {:d} ps \t Channel: {:d} \t MissedEvents: {:d} \t EventType: {:s}'.format(
        len(timestamps), timestamps[-1], channel[-1], missed_events[-1], eventName[overflow_types[-1]]))

print('--- The end of the data is reached. ---')

print("""
The raw time tag stream can be also accessed in real-time using the TimeTagStream class.
The TimeTagStream class returns the tags received directly from the hardware for on-the-fly processing.
""")

print('===============================')
print('Deleting the temporary files.')
del file_reader
tempdir.cleanup()
