from TimeTagger import setLogger, createTimeTagger, Combiner, Coincidence, Counter, Countrate, Correlation, TimeDifferences, TimeTagStream, Scope, Event, CHANNEL_UNUSED, UNKNOWN, LOW, HIGH, LOGGER_WARNING
from pylab import *


def log_handler(level, msg):
    if level >= LOGGER_WARNING:
        print(msg)


# activate trivial custom logging
setLogger(log_handler)

# create a timetagger instance
tagger = createTimeTagger()
tagger.reset()

# simple counter on channels 0 and 1 with 1 ms binwidth (1e9 ps) and 2000 successive values
print("****************************************")
print("*** Demonstrate a counter time trace ***")
print("****************************************")
print("")
print("Create counters on channels 1 and 2 with 1 ms binwidth and 1000 points.")
print("")
count = Counter(tagger, channels=[1, 2], binwidth=int(1e9), n_values=int(1000))


# apply the built-in test signal (~0.8 to 0.9 MHz) to channels 1 and 2
print("Enabling test signal on channel 1.")
print("")
tagger.setTestSignal(1, True)
pause(.5)
print("Enabling test signal on channel 2.")
print("")
tagger.setTestSignal(2, True)
pause(.5)

# wait until the 1000 values should be filled
# we should see the calibration signal turning on in the time trace

# retrieve the current buffer
data = count.getData()

# plot the result
figure()
plot(count.getIndex()/1e12, data[0]*1e-3, label='channel 1')
plot(count.getIndex()/1e12, data[1]*1e-3, label='channel 2')
xlabel('Time [s]')
ylabel('Countrate [MHz]')
ylim(0., 4.)
legend()
title('Time trace of the click rate on channel 1 and 2')
text(0.1, 2, 'The built-in test signal (~ 800 to 900 kHz) is applied first to channel 1\nand 0.5 s later to channel 2.')
tight_layout()
show()

# cross correlation between channels 1 and 2
# binwidth=1000 ps, n_bins=3000, thus we sample 3000 ns
# we should see the correlation delta peaks at a bit more than 1000 ns distance
print("*************************************")
print("*** Demonstrate cross correlation ***")
print("*************************************")
print("")
print("Create a cross correlation on channels 1 and 2 with 1 ns binwidth and 3000 points.")
print("")
corr = Correlation(tagger, channel_1=1, channel_2=2,
                   binwidth=1000, n_bins=3000)
pause(1)
figure()
plot(corr.getIndex()/1e3, corr.getData())
xlabel('Time [ns]')
ylabel('Clicks')
title('Cross correlation between channel 1 and 2')
text(-1400, 300000, """
The built-in test signal is applied
to channel 1 and channel 2.
The peak distance corresponds to
the period of the test signal.
""")
text(100, 300000, """
Note: the decreasing peak heights
and broadening of the peaks
reflects the jitter of the built-in
test signal, which is much larger
than the instrument jitter.
""")
tight_layout()
show()

# cross correlation between channels 1 and 2
# binwidth=10 ps, n_bins=400, thus we sample 4 ns
# The standard deviation of the peak
# is the root mean square sum of the
# input jitters of channels 1 and 2
print("Create a cross correlation on channels 1 and 2 with 10 ps binwidth and 100 points.")
print("")
corr = Correlation(tagger, channel_1=1, channel_2=2, binwidth=10, n_bins=400)
pause(2)
figure()
plot(corr.getIndex(), corr.getData())
title('High res cross correlation showing the instrument jitter')
text(-1500, 80000, """
The half width of the peak is
sqrt(2) times the instrument jitter.
The shift of the peak from zero
is the propagation delay of the
built-in test signal.
""")
xlabel('Time [ps]')
ylabel('Clicks')
tight_layout()
show()

propagation_delay = int(abs(corr.getIndex()[corr.getData().argmax()]))

print("************************************")
print("*** Demonstrate virtual channels ***")
print("************************************")
print("")
print("Create a virtual channel that contains all tags of channel 1 and channel 2.")
print("")
ch1_or_ch2 = Combiner(tagger, channels=[1, 2])
print("The virtual channel was assigned the channel number %i." %
      ch1_or_ch2.getChannel())
print("")
print("Create a virtual channel that contains coincidence clicks of channel 1 and channel 2 within a +-%i ps window." % propagation_delay)
print("")
ch1_and_ch2 = Coincidence(
    tagger, channels=[1, 2], coincidenceWindow=propagation_delay)
print("The virtual channel was assigned the channel number %i." %
      ch1_and_ch2.getChannel())
print("")
print("Create a countrate on channel 1 and the virtual channels.")
print("")
countrate = Countrate(
    tagger, channels=[1, ch1_or_ch2.getChannel(), ch1_and_ch2.getChannel()])
pause(1)
print("Count rates\n  channel 1:    %i counts / s\n  ch1 and ch2: %i counts / s\n  coincidences: %i counts / s." %
      tuple(countrate.getData()))
print("""
Here, we have used a coincidence window that coincides just with the propagation delay
between ch1 and ch2. Due to jitter, approximately half of the coincidence counts
fall outside of this coincidence range. Therefore, the coincidence countrate is
approximately half of the original count rate.
""")

print("***********************************")
print("*** Demonstrate synchronization ***")
print("***********************************")
print("")
print("Disable the test signal.")
tagger.setTestSignal(1, False)
tagger.setTestSignal(2, False)
print("We compare the total counts from a channel with the sync and without.")
print("")
countrate = Countrate(tagger, channels=[1])
pause(1)
print("Clear the data, enable the test signal, wait 1 s, read the data.")
print("")
countrate.clear()
tagger.setTestSignal(1, True)
pause(1)
counts_no_sync = countrate.getCountsTotal()
print("Disable the test signal.")
print("")
tagger.setTestSignal(1, False)
print("Clear the data, enable the test signal, wait until the settings are applied, wait 1 s, read the data.")
print("")
countrate.clear()
tagger.setTestSignal(1, True)
tagger.sync()
pause(1)
counts_sync = countrate.getCountsTotal()
print("Counts\n  without sync: %i counts\n  with sync:    %i counts" %
      (counts_no_sync[0], counts_sync[0]))
print("""
The execution of the 'sync' method ensures that all settings are actually
applied (in this case the activation of test signal) and the USB buffer is
flushed. I.e., after a 'sync' call, we are sure not to receive time tags that
were generated in the past. In this demonstration, we turn on the test signal
and record the total amount of clicks. Data transmission is never instantaneous 
and therefore in the first example, the test signal was turned on but there is
a delay until the first test signal arrives from the hardware over USB.
Consequently, the counts are reduced compared to the actual value. By inserting a 
"sync" call, the behavior is changed and we ensure that the settings are applied 
and the test signal is coming in.
""")

print("*****************************")
print("*** Demonstrate filtering ***")
print("*****************************")
print("")
print("Enabling event filter.")
tagger.setConditionalFilter(trigger=[1], filtered=[5])
print("")
print("Enabling test signal on channel 1 and channel 5.")
print("")
tagger.setTestSignal(1, True)
tagger.setTestSignal(5, True)
rate = Countrate(tagger, channels=[1, 5])
rate.clear()
pause(1)
print("Count rates\n  channel 1: %i counts / s\n  channel 5: %i counts / s" %
      tuple(rate.getData()))
print("")
print("Disabling test signal on channel 1.")
print("")
tagger.setTestSignal(1, False)
rate.clear()
pause(1)
print("Count rates\n  channel 1: %i counts / s\n  channel 5: %i counts / s" %
      tuple(rate.getData()))
print("""
Here, we have used the event filter between channel 1 and channel 5 to suppress 
time tags on channel 5. The filter drops time tags on channel 5 unless they were
preceded by a tag on channel 1. First, since the tags are simultaneous on both
channels, but there is a finite jitter the count rate on channel 5 is reduced by
some amount. Now, as we disable the test signal on channel 1, all time tags on 
channel 5 are suppressed and a count rate of 0 is measured, even though the test
signal is active. The filtering is useful to tame signals with data rates exceeding
the USB transfer rate such as synchronization signals from pulsed lasers in 
fluorescence lifetime measurements.
""")
tagger.clearConditionalFilter()

# disable the internal test signal on channels 1 and 2
print("Disable test signals on channel 1 and 2.")
print("")
tagger.setTestSignal(1, False)
tagger.setTestSignal(2, False)

print("*****************************************")
print("*** Demonstrate trigger level control ***")
print("*****************************************")
print("")
print("Disable test signals and filter.")
print("")
tagger.setTestSignal(1, False)
tagger.setTestSignal(2, False)
tagger.setTestSignal(5, False)
# set the trigger levels on channel 1 and 2 to 0.5 V
# (this is the default value at startup)
print("Set trigger levels on channel 1 and 2 to 0.5 volts.")
print("")
tagger.setTriggerLevel(1, 0.5)
tagger.setTriggerLevel(2, 0.5)
pause(1)

print("**************************************")
print("*** Access the raw time tag stream ***")
print("**************************************")
print("")
print("Enable test signals.")
print("")
tagger.setTestSignal([1, 2], True)
# capture the incoming tags (with a maximum of 1M tags)
stream = TimeTagStream(tagger, n_max_events=1000000, channels=[1, 2])
# take data for 0.5s
pause(0.5)
buffer = stream.getData()
timestamps = buffer.getTimestamps()
channels = buffer.getChannels()
print("Total number of tags stored in the buffer: " + str(buffer.size))
print("Show the first 10 tags")
for i in range(min(buffer.size, 10)):
    print("  time in ps: " +
          str(timestamps[i]) + " signal received on channel: " + str(channels[i]))
print("")

pause(1)

print("*****************************")
print("*** Digital scope example ***")
print("*****************************")
print("")
print("Enable test signals.")
print("")
channel = 1
tagger.setTestSignal(channel, True)
scope = Scope(tagger, event_channels=[
              channel], trigger_channel=channel, window_size=10000000, n_traces=1, n_max_events=1000)
pause(0.5)
edges = scope.getData()

# number of edges received
edgesSize = size(edges[0][:])

print("Received " + str(edgesSize) + " edges for channel 1.")
print("")

if (edgesSize > 0) & (edges[0][0].state != UNKNOWN):
    # every edge is represented as two x,y points
    x = [0] * edgesSize * 2
    y = [0] * edgesSize * 2
    for i in range(edgesSize):
        x[i*2] = edges[0][i].time
        y[i*2] = edges[0][i].state == LOW
        x[i*2+1] = edges[0][i].time
        y[i*2+1] = edges[0][i].state == HIGH

    figure()
    plot(array(x)/1000000., y)
    title('Digital Scope - trace of channel 1 (test signal)')
    ylim(-0.1, 1.1)
    xlabel('Time [us]')
    ylabel('Signal')
    gca().set_yticks([0, 1])
    show()

print("************************")
print("*** startFor example ***")
print("************************")
print("""
With the startFor method, the measurement time can be set with a parameter.
startFor clears the current data (optionally) and stops the measurement after
the time passed as a parameter (here 3s).
""")
tagger.setTestSignal([1, 2], True)

corr = Correlation(tagger, channel_1=1, channel_2=2, binwidth=10, n_bins=400)
corr.startFor(3 * 1000000000000)
while (corr.isRunning()):
    pause(0.4)
    print("Current acquisition time: " +
          str(double(corr.getCaptureDuration())/1e12) + " s")
figure()
plot(corr.getIndex()/1e3, corr.getData())
xlabel('Time [ns]')
ylabel('Clicks')
title('Cross correlation aquired for 3 seconds.')
tight_layout()
show()
