"""
This example shows how to use Virtual Time Tagger for reprocessing earlier stored timetag stream data.

The example consists of two steps.
Step 1:
    Use the real Time Tagger and record timetag data from two channels with test signal enabled.
    In addition, accumulate a coarse histogram of the time differences.
Step 2:
    Use Virtual Time Tagger to reprocess earlier stored data and accumulate time difference 
    histogram with much better time resolution. In addition, measure the countrate on both channels.
"""

import os
import tempfile
import matplotlib.pyplot as plt
from time import sleep
from TimeTagger import *

# File for storing time-tags
data_file = 'example-file.ttbin'

# Temporary file path
tempdir = tempfile.TemporaryDirectory()
tmpfile = tempdir.name + os.sep + data_file

# Channel definitions
CHAN_A = 1
CHAN_B = 2

print("""
****************************************************
STEP 1: Write events from two channels into a file. 
****************************************************
""")

tt = createTimeTagger()

# Enable internal test signal
tt.setTestSignal(CHAN_A, True)
tt.setTestSignal(CHAN_B, True)

# Create Correlation and FileWriter
corr = Correlation(tt, channel_1=CHAN_A, channel_2=CHAN_B,
                   binwidth=20, n_bins=200)
fw = FileWriter(tt, filename=tmpfile, channels=[CHAN_A, CHAN_B])

# Accumulate data for some time
print('Writing data to a file:', tmpfile)
print(' ...')
sleep(10)

# Stop data accumulation
corr.stop()
fw.stop()

# Print statistics on file size and data-rate
print('Data recording complete.')
print('*************************** STATISTICS *****************************')
print('Recording duration:            {:0.0f} seconds'.format(
    fw.getCaptureDuration()*1e-12))
print('Events written to the file(s): {:0.0f} events'.format(
    fw.getTotalEvents()))
print('Bytes written to the file(s):  {:0.0f} bytes'.format(fw.getTotalSize()))
print('File writing data rate:        {:0.0f} bytes/s'.format(
    fw.getTotalSize()/(fw.getCaptureDuration()*1e-12)))
print('Compression efficiency:        {:0.2f} bytes/event'. format(
    fw.getTotalSize()/fw.getTotalEvents()))
print('NOTE: Compression efficiency depends on signal randomness.')
print('********************************************************************')

# Delete FileWriter object
del fw


# Plot the correlation histogram
fig = plt.figure()
ax = fig.subplots()
ax.bar(corr.getIndex(), corr.getData(), 20)
ax.set_xlabel('time (ps)')
ax.set_ylabel('counts')
ax.set_title('Analyzed with large bin size')
fig.tight_layout()
fig.show()

# Free the time tagger resources
freeTimeTagger(tt)

print("""
***********************************************************
STEP 2: Use Virtual Time Tagger to read the data file 
        and repeat the measurement with different binwidth
        In addition, measure count rate on both channels.
***********************************************************
""")

# Create Virtual Time Tagger
ttv = createTimeTaggerVirtual()

# Create the measurements the same way as you do with the real Time Tagger
corr = Correlation(ttv, channel_1=CHAN_A, channel_2=CHAN_B,
                   binwidth=1, n_bins=4000)
crate = Countrate(ttv, channels=[CHAN_A, CHAN_B])

# Specify the file to read the time tags from file or file sequence
ttv.replay(tmpfile)
print("Virtual Time Tagger is streaming time tags from the file...")

# Wait until the file reading is finished
ttv.waitForCompletion()
print("File streaming completed.")

# Create figure and axes
fig = plt.figure()
ax = fig.subplots()
ax.bar(corr.getIndex(), corr.getData(), 1)
ax.set_xlabel('time (ps)')
ax.set_ylabel('counts')
ax.set_title('Reprocessed using TimeTaggerVirtual with smaller binning')
fig.show()
plt.show()

# Free Virtual Time Tagger resources
freeTimeTagger(ttv)

# Delete temporary file
print("""
**********************************************
END: Delete temporary data file.
**********************************************
""")
tempdir.cleanup()
