import matplotlib.pyplot as plt
import TimeTagger
import numpy
import numba

class CustomStartStop(TimeTagger.CustomMeasurement):
    """
    Example single start - multiple stop measurement.
    """

    def __init__(self, tagger, click_channel, start_channel, binwidth, bins):
        TimeTagger.CustomMeasurement.__init__(self, tagger)
        self.click_channel = click_channel
        self.start_channel = start_channel
        self.binwidth = binwidth
        self.max_bins = bins

        # Each used channel must be registered
        self.register_channel(click_channel)
        self.register_channel(start_channel)

        self.clear_impl()

        # At the end of a Measurement construction we must indicate that we
        # have finished
        self.finalize_init()

    def __del__(self):
        # The measurement must be stopped before deconstruction to avoid
        # concurrent measure() calls
        self.stop()

    def getData(self):
        # lock this instance to avoid conflicting results while measure is
        # running apart.
        self._lock()
        arr = self.data.copy()
        # We have gathered the data, unlock so measuring can continue
        self._unlock()
        return arr

    def getIndex(self):
        # this method does not depend on the internal state, so there is no
        # need for a lock
        arr = numpy.arange(0, self.max_bins) * self.binwidth
        return arr

    def clear_impl(self):
        # the lock is already aquired
        self.last_start_timestamp = 0
        self.data = numpy.zeros((self.max_bins,), dtype=numpy.uint64)

    def on_start(self):
        # the lock is already aquired
        pass

    def on_stop(self):
        # the lock is already aquired
        pass

    @staticmethod
    @numba.jit(nopython=True, nogil=True)
    def fast_measure(
            tags,
            data,
            click_channel,
            start_channel,
            binwidth,
            last_start_timestamp):
        """
        A recompiled version of the histogram algorithm for better performance
        nopython=True: Only a subset of the python syntax is supported.
                       Avoid everything but primitives and numpy arrays.
                       All slow operation will yield an exception
        nogil=True:    This method will release the global interpreter lock. So
                       this method can run in parallel with other python code
        """
        for tag in tags:
            # tag.type can be: 0 - TimeTag, 1- Error, 2 - OverflowBegin, 3 -
            # OverflowEnd, 4 - MissedEvents
            if tag['type'] != 0:
                # tag is not a TimeTag
                last_start_timestamp = 0
            elif tag['channel'] == click_channel and last_start_timestamp != 0:
                # valid click
                index = (tag['time'] - last_start_timestamp) // binwidth
                if index < data.shape[0]:
                    data[index] += 1
            if tag['channel'] == start_channel:
                last_start_timestamp = tag['time']
        return last_start_timestamp

    def measure(self, incoming_tags, begin_time, end_time):
        # the lock is already aquired
        # self.data is provided as reference, so it must not be accessed
        # anywhere else without locking the mutex.
        # incoming_tags is provided as a read-only reference. The storage will
        # be deallocated after this call, so you must not store a reference to
        # this object. Make a copy instead.
        self.last_start_timestamp = CustomStartStop.fast_measure(
            incoming_tags,
            self.data,
            self.click_channel,
            self.start_channel,
            self.binwidth,
            self.last_start_timestamp)

# Channel definitions
CHAN_A = 1
CHAN_B = 2

if __name__ == '__main__':

    tt = TimeTagger.createTimeTagger()
    tt.reset() # Reset all settings to default values

    tt.setTestSignal([CHAN_A, CHAN_B], True)
    tt.sync()

    time = 1000 # 1000 picoseconds
    st = CustomStartStop(tt, CHAN_A, CHAN_B, binwidth=time, bins=100000)

    print("Waiting...")
    plt.pause(3.0)

    result = st.getData()

    i = numpy.arange(result.shape[0])
    i = i[result != 0]
    start = i[0]
    end = i[-1]
    x = numpy.arange(start,end) * time
    y = result[start:end]

    plt.plot(x,y)
    plt.show()
